print("import: 'gala'")
import gala

print("import: 'gala.coordinates'")
import gala.coordinates

print("import: 'gala.dynamics'")
import gala.dynamics

print("import: 'gala.dynamics._genfunc'")
import gala.dynamics._genfunc

print("import: 'gala.dynamics.lyapunov'")
import gala.dynamics.lyapunov

print("import: 'gala.dynamics.mockstream'")
import gala.dynamics.mockstream

print("import: 'gala.integrate'")
import gala.integrate

print("import: 'gala.integrate.cyintegrators'")
import gala.integrate.cyintegrators

print("import: 'gala.integrate.cyintegrators.dopri'")
import gala.integrate.cyintegrators.dopri

print("import: 'gala.integrate.pyintegrators'")
import gala.integrate.pyintegrators

print("import: 'gala.potential'")
import gala.potential

print("import: 'gala.potential.frame'")
import gala.potential.frame

print("import: 'gala.potential.frame.builtin'")
import gala.potential.frame.builtin

print("import: 'gala.potential.hamiltonian'")
import gala.potential.hamiltonian

print("import: 'gala.potential.potential'")
import gala.potential.potential

print("import: 'gala.potential.potential.builtin'")
import gala.potential.potential.builtin

print("import: 'gala.potential.scf'")
import gala.potential.scf

